@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134, San Rafael,
-   CA  94903, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
+   CA 94945, U.S.A., +1(415)492-9861, for further information.
 */
 
 
@@ -18,6 +18,8 @@
 
 #include "ghostxps.h"
 
+#define CFF_ARGS_SIZE 48
+
 /*
  * Big-endian memory accessor functions
  */
@@ -171,13 +173,15 @@ xps_read_cff_integer(byte *p, byte *e, int b0, int *val)
 static int
 xps_read_cff_dict(byte *p, byte *e, xps_font_t *font, gs_font_type1 *pt1)
 {
-    struct { int ival; float fval; } args[48];
+    struct { int ival; float fval; } args[CFF_ARGS_SIZE];
     int offset;
     int b0, n;
 
     int privatelen = 0;
     int privateofs = 0;
 
+    memset(args, 0x00, sizeof(args));
+
     offset = p - font->cffdata;
 
     n = 0;
@@ -200,17 +204,25 @@ xps_read_cff_dict(byte *p, byte *e, xps_font_t *font, gs_font_type1 *pt1)
 
             if (b0 == 17)
             {
+                if (args[0].ival < 0)
+                    return gs_throw(-1, "corrupt cff file offset");
                 font->charstrings = font->cffdata + args[0].ival;
             }
 
             if (b0 == 18)
             {
+                if (args[0].ival < 0 || args[1].ival < 0)
+                    return gs_throw(-1, "corrupt cff file offset");
                 privatelen = args[0].ival;
                 privateofs = args[1].ival;
+                if ((font->cffdata + privateofs + privatelen) > font->cffend)
+                    return gs_throw(-1, "corrupt cff file offset");
             }
 
             if (b0 == 19)
             {
+                if (args[0].ival < 0)
+                    return gs_throw(-1, "corrupt cff file offset");
                 font->subrs = font->cffdata + offset + args[0].ival;
             }
 
@@ -340,6 +352,8 @@ xps_read_cff_dict(byte *p, byte *e, xps_font_t *font, gs_font_type1 *pt1)
         {
             if (b0 == 30)
             {
+                if (n >= CFF_ARGS_SIZE)
+                    return gs_throw(-1, "overflow in cff dict");
                 p = xps_read_cff_real(p, e, &args[n].fval);
                 if (!p)
                     return gs_throw(-1, "corrupt dictionary operand");
@@ -348,6 +362,8 @@ xps_read_cff_dict(byte *p, byte *e, xps_font_t *font, gs_font_type1 *pt1)
             }
             else if (b0 == 28 || b0 == 29 || (b0 >= 32 && b0 <= 254))
             {
+                if (n >= CFF_ARGS_SIZE)
+                    return gs_throw(-1, "overflow in cff dict");
                 p = xps_read_cff_integer(p, e, b0, &args[n].ival);
                 if (!p)
                     return gs_throw(-1, "corrupt dictionary operand");
@@ -413,6 +429,12 @@ xps_count_cff_index(byte *p, byte *e, int *countp)
 
     p += count * offsize;
     last = uofs(p, offsize);
+    if (last < 0 || p + last > e)
+    {
+        gs_throw(-1, "corrupt index header");
+        return 0;
+    }
+
     p += offsize;
     p --; /* stupid offsets */
 
@@ -514,7 +536,7 @@ xps_read_cff_file(xps_font_t *font, gs_font_type1 *pt1)
 
     /* CFF header */
     {
-        int major, minor, hdrsize, offsize;
+        int major, minor, hdrsize;
 
         if (p + 4 > e)
             return gs_throw(-1, "not enough data for header");
@@ -522,7 +544,7 @@ xps_read_cff_file(xps_font_t *font, gs_font_type1 *pt1)
         major = *p++;
         minor = *p++;
         hdrsize = *p++;
-        offsize = *p++;
+        /*offsize = **/p++;
 
         if (major != 1 || minor != 0)
             return gs_throw(-1, "not a CFF 1.0 file");
@@ -594,14 +616,14 @@ xps_post_callback_encode_char(gs_font *pfont, gs_char chr, gs_glyph_space_t spc)
     int value;
     value = xps_encode_font_char(xf, chr);
     if (value == 0)
-        return gs_no_glyph;
+        return GS_NO_GLYPH;
     return value;
 }
 
-static gs_char
-xps_post_callback_decode_glyph(gs_font *p42, gs_glyph glyph, int ch)
+static int
+xps_post_callback_decode_glyph(gs_font *p42, gs_glyph glyph, int ch, ushort *unicode_return, unsigned int length)
 {
-    return GS_NO_CHAR;
+    return 0;
 }
 
 static int
@@ -705,13 +727,12 @@ xps_post_callback_pop(void *callback_data, fixed *value)
 }
 
 static int
-xps_cff_append(gs_state *pgs, gs_font_type1 *pt1, gs_glyph glyph, int donthint)
+xps_cff_append(gs_gstate *pgs, gs_font_type1 *pt1, gs_glyph glyph, int donthint)
 {
     int code, value;
     gs_type1exec_state cxs;
     gs_glyph_data_t gd;
     gs_type1_state *const pcis = &cxs.cis;
-    gs_imager_state *pgis = (gs_imager_state*)pgs;
     gs_glyph_data_t *pgd = &gd;
     double sbw[4];
     gs_matrix mtx;
@@ -725,10 +746,10 @@ xps_cff_append(gs_state *pgs, gs_font_type1 *pt1, gs_glyph glyph, int donthint)
     gs_matrix_scale(&mtx, 0.001, 0.001, &mtx);
     mtx.tx = 0;
     mtx.ty = 0;
-    gs_matrix_fixed_from_matrix(&pgis->ctm, &mtx);
-    pgis->flatness = 0;
+    gs_matrix_fixed_from_matrix(&pgs->ctm, &mtx);
+    pgs->flatness = 0;
 
-    code = gs_type1_interp_init(&cxs.cis, pgis, pgs->path, NULL, NULL, donthint, 0, pt1);
+    code = gs_type1_interp_init(&cxs.cis, pgs, pgs->path, NULL, NULL, donthint, 0, pt1);
     if (code < 0)
         return gs_throw(code, "cannot init type1 interpreter");
 
@@ -761,7 +782,7 @@ xps_cff_append(gs_state *pgs, gs_font_type1 *pt1, gs_glyph glyph, int donthint)
 }
 
 static int
-xps_post_callback_build_char(gs_show_enum *penum, gs_state *pgs,
+xps_post_callback_build_char(gs_show_enum *penum, gs_gstate *pgs,
         gs_font *pfont, gs_char chr, gs_glyph glyph)
 {
     gs_font_type1 *pt1 = (gs_font_type1*)pfont;
@@ -811,6 +832,7 @@ xps_init_postscript_font(xps_context_t *ctx, xps_font_t *font)
     gs_font_type1 *pt1;
     int cffofs;
     int cfflen;
+    int cffend;
     int code;
 
     /* Find the CFF table and parse it to create a charstring based font */
@@ -821,11 +843,13 @@ xps_init_postscript_font(xps_context_t *ctx, xps_font_t *font)
     if (cffofs < 0)
         return gs_throw(-1, "cannot find CFF table");
 
-    if (cfflen < 0 || cffofs + cfflen > font->length)
+    /* check the table is within the buffer and no integer overflow occurs */
+    cffend = cffofs + cfflen;
+    if (cffend < cffofs || cfflen < 0 || cffend > font->length)
         return gs_throw(-1, "corrupt CFF table location");
 
     font->cffdata = font->data + cffofs;
-    font->cffend = font->data + cffofs + cfflen;
+    font->cffend = font->data + cffend;
 
     font->gsubrs = 0;
     font->subrs = 0;
@@ -918,10 +942,10 @@ xps_init_postscript_font(xps_context_t *ctx, xps_font_t *font)
     pt1->data.nominalWidthX = 0;
 
     pt1->data.BlueFuzz = 1;
-    pt1->data.BlueScale = 0.039625;
+    pt1->data.BlueScale = 0.039625f;
     pt1->data.BlueShift = 7;
     pt1->data.BlueValues.count = 0;
-    pt1->data.ExpansionFactor = 0.06;
+    pt1->data.ExpansionFactor = 0.06f;
     pt1->data.ForceBold = 0;
     pt1->data.FamilyBlues.count = 0;
     pt1->data.FamilyOtherBlues.count = 0;
