@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -17,6 +17,9 @@
 
 #include "ghostxps.h"
 
+#define INITIAL_DASH_SIZE 32
+#define ADDITIVE_DASH_SIZE 1024
+
 char *
 xps_get_real_params(char *s, int num, float *x)
 {
@@ -65,17 +68,21 @@ xps_clip(xps_context_t *ctx)
 void
 xps_fill(xps_context_t *ctx)
 {
-    if (gs_currentopacityalpha(ctx->pgs) < 0.001)
+    if (gs_getfillconstantalpha(ctx->pgs) < 0.001)
         gs_newpath(ctx->pgs);
     else if (ctx->fill_rule == 0) {
         if (gs_eofill(ctx->pgs) == gs_error_Remap_Color){
+            ctx->in_high_level_pattern = true;
             xps_high_level_pattern(ctx);
+            ctx->in_high_level_pattern = false;
             gs_eofill(ctx->pgs);
         }
     }
     else {
         if (gs_fill(ctx->pgs) == gs_error_Remap_Color){
+            ctx->in_high_level_pattern = true;
             xps_high_level_pattern(ctx);
+            ctx->in_high_level_pattern = false;
             gs_fill(ctx->pgs);
         }
     }
@@ -256,7 +263,7 @@ xps_parse_abbreviated_geometry(xps_context_t *ctx, char *geom)
     float smooth_x, smooth_y; /* saved cubic bezier control point for smooth curves */
     int reset_smooth;
 
-    args = xps_alloc(ctx, sizeof(char*) * (strlen(geom) + 2));
+    args = xps_alloc(ctx, sizeof(char*) * (strlen(geom) + 1));
     if (!args) {
         gs_throw(gs_error_VMerror, "out of memory: args.\n");
         return;
@@ -284,8 +291,7 @@ xps_parse_abbreviated_geometry(xps_context_t *ctx, char *geom)
         }
     }
 
-    pargs[0] = s;
-    pargs[1] = 0;
+    *pargs = s;
 
     n = pargs - args;
     i = 0;
@@ -885,7 +891,6 @@ xps_parse_path(xps_context_t *ctx, char *base_uri, xps_resource_t *dict, xps_ite
     float linewidth;
     float miterlimit;
     float samples[XPS_MAX_COLORS];
-    gs_color_space *colorspace;
 
     bool opacity_pushed = false;
     bool uses_stroke = false;
@@ -991,9 +996,21 @@ xps_parse_path(xps_context_t *ctx, char *base_uri, xps_resource_t *dict, xps_ite
     if (stroke_dash_array_att)
     {
         char *s = stroke_dash_array_att;
-        float dash_array[100];
+        float *dash_array;
         float dash_offset = 0.0;
         int dash_count = 0;
+        int dash_mem_count = 0;
+
+        /* Do an initial reasonable allocation. If that
+           runs out, double until we get to a max size
+           and then just add that max each overrun. */
+        dash_array = xps_alloc(ctx, sizeof(float) * INITIAL_DASH_SIZE);
+        if (dash_array == NULL)
+        {
+            gs_throw(gs_error_VMerror, "out of memory: dash_array.\n");
+            return gs_error_VMerror;
+        }
+        dash_mem_count = INITIAL_DASH_SIZE;
 
         if (stroke_dash_offset_att)
             dash_offset = atof(stroke_dash_offset_att) * linewidth;
@@ -1003,7 +1020,24 @@ xps_parse_path(xps_context_t *ctx, char *base_uri, xps_resource_t *dict, xps_ite
             while (*s == ' ')
                 s++;
             if (*s) /* needed in case of a space before the last quote */
+            {
+                /* Double up to a max size of ADDITIVE_DASH_SIZE and then add
+                   that amount each time */
+                if (dash_count > (dash_mem_count - 1))
+                {
+                    if (dash_mem_count < ADDITIVE_DASH_SIZE)
+                        dash_mem_count = dash_mem_count * 2;
+                    else
+                        dash_mem_count = dash_mem_count + ADDITIVE_DASH_SIZE;
+                    dash_array = (float*) xps_realloc(ctx, dash_array, sizeof(float) * dash_mem_count);
+                    if (dash_array == NULL)
+                    {
+                        gs_throw(gs_error_VMerror, "out of memory: dash_array realloc.\n");
+                        return gs_error_VMerror;
+                    }
+                }
                 dash_array[dash_count++] = atof(s) * linewidth;
+            }
             while (*s && *s != ' ')
                 s++;
         }
@@ -1017,8 +1051,8 @@ xps_parse_path(xps_context_t *ctx, char *base_uri, xps_resource_t *dict, xps_ite
             if (phase_len == 0)
                 dash_count = 0;
         }
-
         gs_setdash(ctx->pgs, dash_array, dash_count, dash_offset);
+        xps_free(ctx, dash_array);
     }
     else
     {
@@ -1067,6 +1101,8 @@ xps_parse_path(xps_context_t *ctx, char *base_uri, xps_resource_t *dict, xps_ite
     }
     if (fill_att)
     {
+        gs_color_space *colorspace;
+
         if (data_att)
             xps_parse_abbreviated_geometry(ctx, data_att);
         if (data_tag)
@@ -1084,6 +1120,7 @@ xps_parse_path(xps_context_t *ctx, char *base_uri, xps_resource_t *dict, xps_ite
         if (fill_opacity_att)
             samples[0] *= atof(fill_opacity_att);
         xps_set_color(ctx, colorspace, samples);
+        rc_decrement(colorspace, "xps_parse_path");
 
         opacity_pushed = true;
         xps_fill(ctx);
@@ -1117,6 +1154,8 @@ xps_parse_path(xps_context_t *ctx, char *base_uri, xps_resource_t *dict, xps_ite
 
     if (stroke_att)
     {
+        gs_color_space *colorspace;
+
         if (data_att)
             xps_parse_abbreviated_geometry(ctx, data_att);
         if (data_tag)
@@ -1137,6 +1176,7 @@ xps_parse_path(xps_context_t *ctx, char *base_uri, xps_resource_t *dict, xps_ite
         if (stroke_opacity_att)
             samples[0] *= atof(stroke_opacity_att);
         xps_set_color(ctx, colorspace, samples);
+        rc_decrement(colorspace, "xps_parse_path");
 
         gs_stroke(ctx->pgs);
     }
